----------------------------------------------------------------------------------------
-- UTIA AV CR, v.v.i
-- Signal Processing Department
-- Pod Vodarenskou vezi 4
-- 182 08 Prague
-- Czech Republic
-- web: http://zs.utia.cas.cz
-- contact: sp@utia.cz
--
-- Filename: axis_vid_det.vhd
-- Date: 09.20.2022
-- Author: Lukas Kohout <kohoutl@utia.cas.cz>
-- Description:
--    resolution detector on AXIS
--
-- Changelog:
--
--
--
--
----------------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

------------------------------------------------------------------------------
------------------------------------------------------------------------------
entity axis_vid_det is
    generic (
        C_TDATA_WIDTH   : natural := 24;
        C_PASS_THROUGH  : natural range 0 to 1 := 1;
        C_X_CNT         : natural range 8 to 4096 := 8;
        C_Y_CNT         : natural range 8 to 4096 := 8;
        C_FRAME_CNT     : natural range 3 to 60 := 3
    );
    port (
        aclk               : in  std_logic;

        aresetn              : in  std_logic;

        -- input AXIS
        s_axis_tvalid     : in  std_logic;
        s_axis_tready     : out std_logic;
        s_axis_tdata      : in  std_logic_vector(C_TDATA_WIDTH-1 downto 0);
        s_axis_tlast      : in  std_logic;
        s_axis_tuser      : in  std_logic_vector(0 downto 0);

        -- output AXIS
        m_axis_tvalid     : out std_logic;
        m_axis_tready     : in  std_logic;
        m_axis_tdata      : out std_logic_vector(C_TDATA_WIDTH-1 downto 0);
        m_axis_tlast      : out std_logic;
        m_axis_tuser      : out std_logic_vector(0 downto 0);

        -- detector output
        x                 : out natural range 0 to C_X_CNT;
        y                 : out natural range 0 to C_Y_CNT;
        valid             : out std_logic

    );

end entity axis_vid_det;

------------------------------------------------------------------------------
architecture rtl of axis_vid_det is


------------------------------------------------------------------------------


    signal s_x_cnt        : natural range 0 to C_X_CNT;
    signal s_y_cnt        : natural range 0 to C_Y_CNT;

--     signal s_x_cnt_ff     : natural range 0 to C_X_CNT;
--     signal s_y_cnt_ff     : natural range 0 to C_Y_CNT;

    type t_x_mem is array (natural range <>) of natural range 0 to C_X_CNT;
    type t_y_mem is array (natural range <>) of natural range 0 to C_Y_CNT;

    signal s_x_mem        : t_x_mem(C_FRAME_CNT-1 downto 0);
    signal s_y_mem        : t_y_mem(C_FRAME_CNT-1 downto 0);

    signal s_sof          : std_logic;
    signal s_sof_ff       : std_logic;
    signal s_eol          : std_logic;
    signal s_active       : std_logic;


    signal s_x_valid      : std_logic;
    signal s_y_valid      : std_logic;


------------------------------------------------------------------------------
begin

------------------------------------------------------------------------------
-- AXIS

    g_pass_trough : if C_PASS_THROUGH = 1 generate

        s_axis_tready <= m_axis_tready;

        m_axis_tvalid <= s_axis_tvalid;
        m_axis_tdata  <= s_axis_tdata;
        m_axis_tlast  <= s_axis_tlast;
        m_axis_tuser  <= s_axis_tuser;

    end generate g_pass_trough;

    g_end_point : if C_PASS_THROUGH = 0 generate

        s_axis_tready <= '1';

    end generate g_end_point;


------------------------------------------------------------------------------
    s_active <= m_axis_tready and s_axis_tvalid;

    s_sof <= s_axis_tuser(0) and s_active;
    s_eol <= s_axis_tlast    and s_active;


------------------------------------------------------------------------------
-- input register
    p_ffs : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            s_sof_ff  <= s_sof;
        end if;
    end process p_ffs;


------------------------------------------------------------------------------
-- line counter
    p_line_cnt : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            if (aresetn = '0') or (s_sof = '1') then
                s_y_cnt <= 0;
            elsif s_eol = '1' then
                s_y_cnt <= s_y_cnt + 1;
            end if;
        end if;
    end process p_line_cnt;

--     p_y_cnt_prev : process(aclk) is
--     begin
--         if aclk'event and aclk = '1' then
--             if aresetn = '0' then
--                 s_y_cnt_ff <= 0;
--             elsif s_sof = '1' then
--                 s_y_cnt_ff <= s_y_cnt;
--             end if;
--         end if;
--     end process p_y_cnt_prev;


------------------------------------------------------------------------------
-- pixel counter
    p_pix_cnt : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            if (aresetn = '0') or (s_eol = '1') then
                s_x_cnt <= 0;
            elsif s_active = '1' then
                s_x_cnt <= s_x_cnt + 1;
            end if;
        end if;
    end process p_pix_cnt;


--     p_x_cnt_prev : process(aclk) is
--     begin
--         if aclk'event and aclk = '1' then
--             if aresetn = '0' then
--                 s_x_cnt_ff <= 0;
--             elsif s_eol = '1' then
--                 s_x_cnt_ff <= s_x_cnt;
--             end if;
--         end if;
--     end process p_x_cnt_prev;


------------------------------------------------------------------------------
-- mem
    p_mem_x : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            if aresetn = '0' then
                s_x_mem <= (0 => 1, others => 0);
            elsif s_eol = '1' then
                s_x_mem <= s_x_mem(C_FRAME_CNT-2 downto 0) & s_x_cnt;
            end if;
        end if;
    end process p_mem_x;

    p_mem_y : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            if (aresetn and s_x_valid) = '0' then
--             if (aresetn) = '0' then
                s_y_mem <= (0 => 1, others => 0);
            elsif s_sof = '1' then
                s_y_mem <= s_y_mem(C_FRAME_CNT-2 downto 0) & s_y_cnt;
            end if;
        end if;
    end process p_mem_y;


------------------------------------------------------------------------------
-- valid
    p_x_valid : process (aclk) is
    variable v_x_valid_mem : std_logic_vector(C_FRAME_CNT - 2 downto 0) := (others => '0');
    variable v_x_valid : std_logic := '0';
    begin
        if aclk'event and aclk = '1' then
            if aresetn = '0' then
                s_x_valid <= '0';
                v_x_valid := '0';
                v_x_valid_mem := (others => '0');
--             elsif s_eol = '1' then
            else
                for i in 0 to C_FRAME_CNT - 2 loop
                    if s_x_mem(i) = s_x_mem(i+1) then
                        v_x_valid_mem(i) := '1';
                    else
                        v_x_valid_mem(i) := '0';
                    end if;
                end loop;

                v_x_valid := '1';
                for i in 0 to C_FRAME_CNT - 2 loop
                    v_x_valid := v_x_valid and v_x_valid_mem(i);
                end loop;

                s_x_valid <= v_x_valid;

            end if;
        end if;

    end process p_x_valid;


    p_y_valid : process (aclk) is
    variable v_y_valid_mem : std_logic_vector(C_FRAME_CNT - 2 downto 0) := (others => '0');
    variable v_y_valid : std_logic := '0';
    begin
        if aclk'event and aclk = '1' then
            if (aresetn and s_x_valid) = '0' then
                s_y_valid <= '0';
                v_y_valid := '0';
                v_y_valid_mem := (others => '0');
            else
                for i in 0 to C_FRAME_CNT - 2 loop
                    if s_y_mem(i) = s_y_mem(i+1) then
                        v_y_valid_mem(i) := '1';
                    else
                        v_y_valid_mem(i) := '0';
                    end if;
                end loop;

                v_y_valid := '1';
                for i in 0 to C_FRAME_CNT - 2 loop
                    v_y_valid := v_y_valid and v_y_valid_mem(i);
                end loop;


                s_y_valid <= v_y_valid;

            end if;
        end if;

    end process p_y_valid;


------------------------------------------------------------------------------
-- output
    p_out : process (aclk) is
    begin
        if aclk'event and aclk = '1' then
            if aresetn = '0' then
                x <= 0;
                y <= 0;
            elsif (s_x_valid and s_y_valid) = '1' then
--                 if s_y_cnt = s_y_cnt_ff then
                    x <= s_x_mem(C_FRAME_CNT-1) + 1;
                    y <= s_y_mem(C_FRAME_CNT-1);
--                 else
--                     x <= 0;
--                     y <= 0;
--                 end if;
            else
                x <= 0;
                y <= 0;
            end if;
        end if;
    end process p_out;

    valid <= s_sof_ff;

end architecture rtl;

