------------------------------------------------------------------------------
--  UTIA AV CR, v.v.i
--  Signal Processing Department
--  Pod Vodarenskou vezi 4
--  182 08 Prague
--  Czech Republic
--  web: http://zs.utia.cas.cz
--  contact: sp@utia.cz
--
--  Filename: im_hdmi_out.vhd
--  Date: 15.09.2016
--  Author: Lukas Kohout <kohoutl@utia.cas.cz>
--  Description:
--     HDMI out for AVNET FMC board
--
--  Changelog:
--    12.04.2017 - UltraScale/UltraScale+ support
--
--
--
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all;

------------------------------------------------------------------------------
entity im_hdmi_out is
    generic (
        C_DATA_WIDTH  : natural := 16;
        C_CLAMP_Y     : natural range 0 to 1 := 1;
        C_CLAMP_CRCB  : natural range 0 to 1 := 1;
        C_FAMILY      : string := "zynq"
    );
    port (
        vid_clk       : in  std_logic;
        vid_vblank    : in  std_logic;
        vid_hblank    : in  std_logic;
        vid_de        : in  std_logic;
        vid_data      : in  std_logic_vector(C_DATA_WIDTH-1 downto 0);

        oe            : in  std_logic;

        hdmi_data     : out std_logic_vector(C_DATA_WIDTH-1 downto 0);
        hdmi_clk      : out std_logic
    );
end entity im_hdmi_out;

------------------------------------------------------------------------------
architecture rtl of im_hdmi_out is

    component adv7511_embed_syncs is
        port (
            clk               : in  std_logic;
            -- Video Input
            vblank_i          : in  std_logic;
            hblank_i          : in  std_logic;
            de_i              : in  std_logic;
            video_data_i      : in  std_logic_vector(15 downto 0);
            -- Video Output
            video_data_o      : out std_logic_vector(15 downto 0)
        );
    end component adv7511_embed_syncs;

------------------------------------------------------------------------------
    constant C_NET_GND    : std_logic := '0';
    constant C_NET_VCC    : std_logic := '1';

    signal s_vid_vblank   : std_logic;
    signal s_vid_hblank   : std_logic;
    signal s_vid_de       : std_logic;
    signal s_vid_data     : std_logic_vector((C_DATA_WIDTH-1) downto 0);
    signal s_vid_y        : std_logic_vector(((C_DATA_WIDTH/2)-1) downto 0);
    signal s_vid_crcb     : std_logic_vector(((C_DATA_WIDTH/2)-1) downto 0);

    signal s_vid_data_es  : std_logic_vector((C_DATA_WIDTH-1) downto 0);
   
--     signal s_oen          : std_logic;
   
begin


------------------------------------------------------------------------------
-- clamp input data from full range (0-255) to 1-254 because of embedded syncs

    clamp_y: if C_CLAMP_Y = 1 generate
        p_clamp_y : process (vid_clk) is
        variable v_y : std_logic_vector(7 downto 0);
        begin
            if vid_clk'event and vid_clk = '1' then

                -- vid_data[15: 8] => Y (luma)
                -- vid_data[ 7: 0] => CbCr (chroma)
                v_y := vid_data(15 downto 8);

                -- Y clamping
                if ( v_y = X"00" ) then
                    s_vid_y <= X"01";
                elsif ( v_y = X"FF" ) then
                    s_vid_y <= X"FE";
                else
                    s_vid_y <= v_y;
                end if;

            end if;
        end process p_clamp_y;
    end generate clamp_y;

    no_clamp_y: if C_CLAMP_Y = 0 generate
        p_no_clamp_y : process (vid_clk) is
        begin
            if vid_clk'event and vid_clk = '1' then

                -- vid_data[15: 8] => Y (luma)
                -- vid_data[ 7: 0] => CbCr (chroma)
                s_vid_y <= vid_data(15 downto 8);

            end if;
        end process p_no_clamp_y;
    end generate no_clamp_y;

    clamp_crcb: if C_CLAMP_CRCB = 1 generate
        p_clamp_crcb : process (vid_clk) is
        variable v_crcb : std_logic_vector(7 downto 0);
        begin
            if vid_clk'event and vid_clk = '1' then

                -- vid_data[15: 8] => Y (luma)
                -- vid_data[ 7: 0] => CbCr (chroma)
                v_crcb := vid_data(7 downto 0);

                -- CrCb clamping
                if ( v_crcb = X"00" ) then
                    s_vid_crcb <= X"01";
                elsif ( v_crcb = X"FF" ) then
                    s_vid_crcb <= X"FE";
                else
                    s_vid_crcb <= v_crcb;
                end if;

            end if;
        end process p_clamp_crcb;
    end generate clamp_crcb;


    no_clamp_crcb: if C_CLAMP_CRCB = 0 generate
        p_no_clamp_crcb : process (vid_clk) is
        begin
            if vid_clk'event and vid_clk = '1' then

                -- vid_data[15: 8] => Y (luma)
                -- vid_data[ 7: 0] => CbCr (chroma)
                s_vid_crcb <= vid_data(7 downto 0);

            end if;
        end process p_no_clamp_crcb;
    end generate no_clamp_crcb;

    s_vid_data <= s_vid_y & s_vid_crcb;


------------------------------------------------------------------------------
-- align syncs with input data
    p_in_ffs : process (vid_clk) is
    begin
        if vid_clk'event and vid_clk = '1' then
            s_vid_vblank  <= vid_vblank;
            s_vid_hblank  <= vid_hblank;
            s_vid_de      <= vid_de;
        end if;
    end process p_in_ffs;



------------------------------------------------------------------------------
-- ADV7511 embedded Syncs
    i_adv7511_embed_syncs : adv7511_embed_syncs
        port map (
            clk           => vid_clk,
            vblank_i      => s_vid_vblank,
            hblank_i      => s_vid_hblank,
            de_i          => s_vid_de,
            video_data_i  => s_vid_data,
            video_data_o  => s_vid_data_es
        );


------------------------------------------------------------------------------
    family_7 : if   ((C_FAMILY = "kintex7") or (C_FAMILY = "virtex7") or (C_FAMILY = "artix7") or (C_FAMILY = "zynq")) generate       

        ------------------------------------------------------------------------------
        -- data
        hdmi_data_buf: for i in s_vid_data_es'low to s_vid_data_es'high generate
            i_ODDR_hdmi_data : ODDR
                generic map (
                    DDR_CLK_EDGE => "SAME_EDGE",  --OPPOSITE_EDGE
                    INIT         => '0',
                    SRTYPE       => "ASYNC"
                )
                port map (
                    Q    => hdmi_data(i),
                    C    => vid_clk,
                    CE   => oe,  --C_NET_VCC,
                    D1   => s_vid_data_es(i),
                    D2   => s_vid_data_es(i),
                    R    => C_NET_GND,
                    S    => C_NET_GND
                );
        end generate hdmi_data_buf;


        ------------------------------------------------------------------------------
        -- pixel clock
        i_ODDR_hdmi_clk : ODDR
            generic map (
                DDR_CLK_EDGE => "SAME_EDGE",
                INIT         => '1',
                SRTYPE       => "ASYNC"
            )
            port map (
                Q    => hdmi_clk,
                C    => vid_clk,
                CE   => oe,
                D1   => C_NET_VCC,
                D2   => C_NET_GND,
                R    => C_NET_GND,
                S    => C_NET_GND
            );

    end generate family_7;


------------------------------------------------------------------------------
    family_8 : if ((C_FAMILY = "kintexu") or (C_FAMILY = "virtexu") or (C_FAMILY = "artixu") or (C_FAMILY = "kintexuplus") or (C_FAMILY = "zynquplus") )   generate

        ------------------------------------------------------------------------------
        -- output enable
--         s_oen  <= not oe;

    
        ------------------------------------------------------------------------------
        -- data
        hdmi_data_buf: for i in s_vid_data_es'low to s_vid_data_es'high generate
            i_ODDRE1_hdmi_data : ODDRE1
                generic map (
                    IS_C_INVERTED   => '0', -- Optional inversion for C
                    IS_D1_INVERTED  => '0', -- Optional inversion for D1
                    IS_D2_INVERTED  => '0', -- Optional inversion for D2
                    SRVAL           => '0'  -- Initializes the ODDRE1 Flip-Flops to the specified value (1’b0, 1’b1)
                )
                port map (
                    Q   => hdmi_data(i),      -- 1-bit output: Data output to IOB
                    C   => vid_clk,           -- 1-bit input: High-speed clock input
                    D1  => s_vid_data_es(i),  -- 1-bit input: Parallel data input 1
                    D2  => s_vid_data_es(i),  -- 1-bit input: Parallel data input 2
                    SR  => C_NET_GND          -- 1-bit input: Active High Async Reset
                );
        end generate hdmi_data_buf;


        ------------------------------------------------------------------------------
        -- pixel clock
        
        hdmi_clk <= vid_clk;
        
--           i_ODDRE1_hdmi_clk : ODDRE1
--               generic map (
--                   IS_C_INVERTED   => '0', -- Optional inversion for C
--                   IS_D1_INVERTED  => '0', -- Optional inversion for D1
--                   IS_D2_INVERTED  => '0', -- Optional inversion for D2
--                   SRVAL           => '0'  -- Initializes the ODDRE1 Flip-Flops to the specified value (1’b0, 1’b1)
--               )
--               port map (
--                   Q   => hdmi_clk,          -- 1-bit output: Data output to IOB
--                   C   => vid_clk,           -- 1-bit input: High-speed clock input
--                   D1  => C_NET_VCC,         -- 1-bit input: Parallel data input 1
--                   D2  => C_NET_GND,         -- 1-bit input: Parallel data input 2
--                   SR  => C_NET_GND          -- 1-bit input: Active High Async Reset
--               );

    end generate family_8;
    
end rtl;
